﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using AForge.Math;

namespace GaCa
{
	class Development
	{
		public enum Interpretation { Decimal, Sum };
		public enum Boundary { Cyclic, Trap };
		public enum Type { NSteps, Transient, OnePeriod, TransientOnePeriod, ThreePeriods, TransientThreePeriods }

		//CASettings settings;
		CA ca;
		Transition transition;
		string init;

		public Development()
		{
		}

		public Development(CA ca)
		{
			this.ca = ca;
			this.transition = new Transition(ca);
		}

		public Phenotype Develop(CA ca, string init)
		{
			this.ca = ca;
			this.init = init;
			//this.settings = settings;
			this.transition = new Transition(ca);

			List<string> states = States(ca);

			//string lastState = states.Last();
			//int prePeriodicLength = states.IndexOf(lastState);
			//states.RemoveAt(states.Count - 1);
			//int periodicLength = states.Count - prePeriodicLength;



			List<int> interpretations = Interpretations(states);

			List<int> magnitudes = Magnitudes(interpretations);
			List<int> halfMagnitudes = magnitudes.GetRange(0, magnitudes.Count / 2);
			List<double> variables = Variables(halfMagnitudes);

			return new Phenotype
			{
				Genotype = ca,
				States = states,
				Interpretations = interpretations,
				//PrePeriodicLength = prePeriodicLength,
				//PeriodicLength = periodicLength,
				Magnitudes = magnitudes,
				//HalfMagnitudes = halfMagnitudes,
				Variables = variables,
			};
        }

		List<string> States(CA ca)
		{
			List<string> states = new List<string>();
			string prev = this.init;
			states.Add(prev);
			for (int i = 0; ; i++)
			{
				string next = Next(prev);
				//int point = states.IndexOf(next);
				if (ca.Settings.DevelopmentType == Type.NSteps)
				{
					if (i >= ca.Settings.Steps)
					{
						return states;
					}
				}
				else
				{
					int point = states.IndexOf(next);
					int limit = 1000;
					if (point != -1 || i >= limit)
					{
						int prePeriodicLength = i < limit ? point : limit;
						int periodicLength = states.Count - prePeriodicLength;
						switch (ca.Settings.DevelopmentType)
						{
							case Type.Transient:
								return states.GetRange(0, prePeriodicLength);
							case Type.OnePeriod:
								return states.GetRange(prePeriodicLength, periodicLength);
							case Type.ThreePeriods:
								return states.GetRange(prePeriodicLength, periodicLength).Multiply(3);
							case Type.TransientOnePeriod:
								return states;
							case Type.TransientThreePeriods:
								List<string> transient = states.GetRange(0, prePeriodicLength);
								List<string> period = states.GetRange(prePeriodicLength, periodicLength);
								return transient.WithReturn(period.Multiply(3));
						}
					}
				}
				states.Add(next);
				prev = next;
			}
		}

		public string Next(string init, int steps)
		{
			string current = init;
			for (int i = 0; i < steps - 1; i++) // including initial!
			{
				string next = Next(current);
				current = next;
			}
			return current;
		}

		public string Next(string state)
		{
            StringBuilder ms = new StringBuilder();
			for (int x = 0; x < ca.Settings.Size; x++)
			{
                StringBuilder conf = new StringBuilder();
                for(int i = - ca.Settings.Neighborhood / 2; i <= ca.Settings.Neighborhood / 2; i++)
                {
                    if (ca.Settings.Boundary == Development.Boundary.Cyclic)
                    {
                        int k = (x + i + ca.Settings.Size) % ca.Settings.Size;
                        conf.Append(state[k]);
                    }
                    else
                    {
                        int k = x + i;
                        if (k < 0 || k >= ca.Settings.Size)
                            conf.Append('0');
                        else
                            conf.Append(state[k]);
                    }
                    
                }
				ms.Append(transition.NextState(conf.ToString()));
			}
			return ms.ToString();
		}

		List<int> Interpretations(List<string> states)
		{
			List<int> interpretations = new List<int>();
			foreach(string state in states)
			{
				if (ca.Settings.Interpretation == Interpretation.Sum)
					interpretations.Add(Sum(state));
				else if (ca.Settings.Interpretation == Interpretation.Decimal)
					interpretations.Add(Decimal(state));
			}
			return interpretations;
		}

        int Decimal(string state)
        {
			int multiplier = (int)(Math.Log10(4000) / Math.Log10(ca.Settings.StatesPerCell));
			ulong scalingFactor = Tools.Pow(ca.Settings.StatesPerCell, state.Length - multiplier);
			ulong interpretation = Tools.BaseToDecimal(state, ca.Settings.StatesPerCell);
			return (int)(interpretation / scalingFactor);
        }

		int Sum(string state)
		{
			double scalingFactor = 0.5 * state.Length / (4000.0 / ca.Settings.StatesPerCell);
			int sum = 0;
			foreach (char c in state)
				sum += (int)Char.GetNumericValue(c);
			return (int)(sum / scalingFactor);
		}

        List<int> Magnitudes(List<int> interpretations)
        {
			//List<int> interpretations = list.ToArray().ToList();
			//List<int> periodicPath = interpretations.GetRange(prePeriodicLength, periodicLength);
			//for (int i = 0; i < numberOfPeriods - 1; i++)
				//interpretations.AddRange(periodicPath);

            Complex[] complexOutputs = Tools.ToComplexes(interpretations);
            FourierTransform.DFT(complexOutputs, FourierTransform.Direction.Forward);
			List<int> magnitudes = Tools.ToMagnitudes(complexOutputs);
			//return magnitudes.GetRange(0, magnitudes.Count / 2);
			return magnitudes;
        }

		List<double> Variables(List<int> halfMagnitudes)
		{
			double scalingFactor = 200;
			List<double> variables = new List<double>();
			foreach (int spikeTime in SpikeTimes(halfMagnitudes))
			{
				double variable = Math.Round((halfMagnitudes[spikeTime] - ca.Settings.Threshold) / scalingFactor, 2);
				variables.Add(variable);
			}
			return variables;
		}

		List<int> SpikeTimes(List<int> points)
		{
			List<int> spikeTimes = new List<int>();
			points.Insert(0, 0);
			points.Add(0);
			for (int x = 1; x < points.Count - 1; x++)
			{
				if (points[x] > ca.Settings.Threshold)
					if (points[x] > points[x - 1] && points[x] > points[x + 1])
						spikeTimes.Add(x);
			}
			return spikeTimes;
		}
           
	}
}
